<?php

namespace WebsiteBundle\Services\Data;

use Bazookas\APIFrameworkBundle\Services\Data\Base\BaseDataService;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Query;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use WebsiteBundle\Entity\Experiment;
use WebsiteBundle\Entity\ExperimentCondition;
use WebsiteBundle\Entity\Participant;
use WebsiteBundle\Entity\Test;
use WebsiteBundle\Entity\TestSession;

class SessionDataService extends BaseDataService
{

  /**
   * @var EntityManager
   */
  private $entityManager;

  public function __construct(EntityManager $entityManager)
  {
    $this->entityManager = $entityManager;
  }

  public function createNewFirstSession(Participant $participant, Experiment $experiment) {
    $condition = $participant->getCondition();
    if (empty($condition)) {
      throw new AccessDeniedException();
    }

    $now = new \DateTime('now', new \DateTimeZone('UTC'));

    $testSession = new TestSession();
    $testSession
      ->setNumber(1)
      ->setParticipant($participant)
      ->setExperiment($experiment)
      ->setScheduled($now)
      ->setAvailableFrom($now)
      //3 sets of trials series 1 61 questions, series 2 and 3 60 questions, 1) 3000ms interval, 2) 2000ms interval, 3) 1500ms interval
      // 393 seconds
      ->setDuration(393)
      //should always start with 3000
      ->setInitialInterval(3000)
      ->setType(Test::TYPE_PASAT)
      ->setGamificationLevel($condition->getGamificationLevel())
    ;

    $this->entityManager->persist($testSession);
    $this->entityManager->flush();

    return $testSession;
  }

  /**
   * @param Participant $participant
   * @param Experiment $experiment
   * @param \DateTime $scheduled
   * @param TestSession $lastSession
   */
  public function createNewSession(Participant $participant, Experiment $experiment, \DateTime $scheduled, TestSession $lastSession)
  {
    //If the user has completed the last session then it is not necessary to create a new session.
    if ($experiment->getTotalSessions() == $lastSession->getNumber()) {
      return;
    }
    $condition = $participant->getCondition();

    $newSession = new TestSession();
    $newSession
      ->setParticipant($participant)
      ->setExperiment($experiment)
      ->setNumber($lastSession->getNumber() + 1)
      ->setScheduled($scheduled)
      ->setAvailableFrom($scheduled)
      ->setDuration($experiment->getSessionDuration())
      ->setInitialLevel($lastSession->getFinalLevel())
    ;

    //if it's not the last session
    if ($newSession->getNumber() != $experiment->getTotalSessions()) {
      $newSession
        ->setType($condition->getTestGroup())
        ->setGamificationLevel($condition->getGamificationLevel())
      ;

      //High gamification games must be assigned a background color
      if ($condition->getGamificationLevel() === ExperimentCondition::GAMIFICATION_HIGH) {
        $newSession->setColor($this->pickRandomColor($lastSession->getColor()));
      }

      //Decide the starting interval for the new session
      if ($lastSession->getNumber() == 1) {
        $entityRepo = $this->entityManager->getRepository(TestSession::class);
        $firstSessionResults = $entityRepo->getFirstSessionResults($participant, $lastSession);
        $firstSessionResults = $this->calculateSessionResults($firstSessionResults);

        $initialInterval = $this->getInitialInterval($firstSessionResults, $condition->getTestGroup());

        $newSession->setInitialInterval($initialInterval);
      } else {
        $newSession->setInitialInterval($lastSession->getMedianInterval() + 500);
      }

    } else {
      //The last session must always be of type PASAT
      $newSession
        ->setType(Test::TYPE_PASAT)
        //should always take 6mins
        ->setDuration(390)
        //should always start with 3000
        ->setInitialInterval(3000)
        ->setGamificationLevel($condition->getGamificationLevel())
      ;
    }

    //Persist the session
    $this->entityManager->persist($newSession);
    $this->entityManager->flush();

    //Increment the session count of the experiment
    $experimentRepo = $this->entityManager->getRepository(Experiment::class);
    $experimentRepo->incrementSessionCount($experiment);
  }

  /**
   * @param array $results
   * @return array
   */
  private function calculateSessionResults(array $results)
  {
    //initialise the data, because in case they got nothing correct the results array will be empty
    $data = [
      3000 => 0,
      2000 => 0,
      1500 => 0,
    ];

    foreach ($results as $result) {
      $data[$result['interval']] = ($result[1] / 60) * 100;
    }

    return $data;
  }

  /**
   * @param array $calculatedResults
   * @param $testType
   * @return int
   */
  private function getInitialInterval(array $calculatedResults, $testType)
  {
    switch ($testType) {
      case Test::TYPE_APASAT:
        return $this->getAPASATInitialInterval($calculatedResults);
      case Test::TYPE_CONTROL:
      default:
        return $this->getControlInitialInterval($calculatedResults);
    }
  }

  /**
   * @param array $calculatedResults
   * @return int
   */
  private function getAPASATInitialInterval(array $calculatedResults)
  {
    if ($calculatedResults[3000] < 51) {
      return 4000;
    }

    if ($calculatedResults[2000] < 51) {
      return 3500;
    }

    if ($calculatedResults[1500] < 51) {
      return 2500;
    }

    return 2000;
  }

  /**
   * @param array $calculatedResults
   * @return int
   */
  private function getControlInitialInterval(array $calculatedResults)
  {
    if ($calculatedResults[3000] < 51) {
      return 3000;
    }

    if ($calculatedResults[2000] < 51) {
      return 3000;
    }

    if ($calculatedResults[1500] < 51) {
      return 2000;
    }

    return 1500;
  }

  private function pickRandomColor($excludeColor) {
    $colors = [
      TestSession::BACKGROUND_COLOR_BLUE,
      TestSession::BACKGROUND_COLOR_PURPLE,
      TestSession::BACKGROUND_COLOR_RED,
      TestSession::BACKGROUND_COLOR_YELLOW
    ];

    $index = array_search($excludeColor, $colors);
    if ($index !== false) {
      unset($colors[$index]);
    }

    return $colors[array_rand($colors)];
  }

  /**
   * Update the scheduled datetime of an existing session (will always be the next incomplete session)
   * @param Participant $participant
   * @param Experiment $experiment
   * @param \DateTime $scheduled
   * @return null|TestSession
   */
  public function updateSession(Participant $participant, Experiment $experiment, \DateTime $scheduled) {
    //Get the session the new session
    $entityRepo = $this->entityManager->getRepository(TestSession::class);
    $nextSession = $entityRepo->getNextSession($participant, $experiment, Query::HYDRATE_OBJECT);

    //Make sure the new time is in UTC
    $scheduled->setTimezone(new \DateTimeZone('UTC'));

    //Update the new session
    $nextSession
      ->setScheduled($scheduled)
    ;

    $this->entityManager->persist($nextSession);
    $this->entityManager->flush();

    return $nextSession;
  }
}
