## Adding menu items

Please note that the menu service configuration described below is only applicable to v2 admin controllers.

### Extending the base menu twig extension

The admin menu is populated by the `BaseAdminMenuExtension` class. In order to add your own menu items you should extend this class and overwrite its service definition. Let's start by extending the class:

```php
<?php

namespace WebsiteBundle\Twig;

use Bazookas\AdminBundle\Twig\Base\BaseAdminMenuExtension;
use Symfony\Component\HttpFoundation\Request;

class AdminMenuExtension extends BaseAdminMenuExtension
{

  protected function createMenu(Request $request) {
    parent::createMenu($request);
    
    //Add your menu items here
  }

}
```

Once you've extended the base class overwrite the `services.yml` definition of the admin menu twig extension:

```yaml
# -----------------------------------------------------------------------------
#                             Twig extensions
# -----------------------------------------------------------------------------
  bazookas.admin.twig.admin_menu_extension:
      class: WebsiteBundle\Twig\AdminMenuExtension  #add your class here
      arguments:
        - "%kernel.bundles%"
        - "%bazookas_admin.menu%"
      tags:
          - { name: twig.extension }
```

Now that you've done the basic setup you can start adding new menu items

### Adding your own menu items

The menu system extension is able to render any kind of AdminElement. However menu specific admin elements have been created.

#### MenuHeaderElement

The menu header element allows you to insert a small piece of header text above a series of menu items.

You can define an element as follows:
```php
new MenuHeaderElement([], [
  'label' => 'admin.menu.admin.header'
]);
```

#### MenuActionElement

The menu action renders a menu action (duh!). This admin elements works in the same way as any other kind of ActionElement. The MenuActionElement accepts the following options in addition to the default admin element properties:

```php
[
  'route' => '',
  'action' => '',
  'label' => '',
  'routeParams' => [],
]
```
You can define a new MenuActionElement as follows:

```php
new MenuActionElement([], [
    'label' => 'admin.entities.cMSUser.namePlural',
    'route' => 'admin_cms_user',
    'iconClass' => 'ti ti-user',
    'roles' => [Roles::ROLE_ADD_USER],
])
```

#### MenuContainer

A menu container allows you to nest several menu items under a single menu item. 

You can create a new menu container as follows:
```php
new MenuContainer([
  new MenuActionElement([], [
    'label' => 'admin.entities.cMSUser.namePlural',
    'route' => 'admin_cms_user',
    'iconClass' => 'ti ti-user',
    'roles' => [Roles::ROLE_ADD_USER],
  ]),
  new MenuActionElement([], [
    'label' => 'admin.entities.cMSUserGroup.namePlural',
    'route' => 'admin_cms_user_group',
    'iconClass' => 'ti-id-badge',
    'roles' => [Roles::ROLE_ADD_GROUP],
  ])
],
[
  'label' => 'admin.menu.users',
  'roles' => [Roles::ROLE_ADD_USER, Roles::ROLE_ADD_GROUP],
  'iconClass' => 'ti ti-user'
]);
```

#### Tips and tricks

The `BaseAdminMenuExtension` class creates the basic menu by invoking a set of functions who are each responsible for creating a small subset of the menu. For example the default implementation of the `createMenu` function you've overwritten above implements:

```php
  /**
   * @param Request $request
   */
  protected function createMenu(Request $request) {
    $this->addMenuHeader();
    $this->addAdminBundleMenuItems();
    $this->addMediaBundleMenuItems();
    $this->addCronBundleMenuItems();
  }
```

As you can see while overriding this function it is easy to add the default menu items in your own function (if you don't want to invoke `parent::createMenu($request);`). This allows you to, for example, change the order in which the items are rendered. Furthermore it is possible to overwrite each of these functions to customise an aspect of the default menu items.

A case where you might want to do this is when you've defined your own CMSUsers. In that case you might want to overwrite the `addAdminBundleMenuItems` function in order to change the labels of the menu items. A sample implementation is shown below:

```php
protected function addAdminBundleMenuItems() {
    //Add the user management menu items
    $this->elements[] = new MenuContainer([
      new MenuActionElement([], [
        'label' => 'admin.entities.researcher.namePlural',
        'route' => 'admin_cms_user',
        'iconClass' => 'ti ti-user',
        'roles' => [Roles::ROLE_ADD_USER],
      ]),
      new MenuActionElement([], [
        'label' => 'admin.entities.cMSUserGroup.namePlural',
        'route' => 'admin_cms_user_group',
        'iconClass' => 'ti-id-badge',
        'roles' => [Roles::ROLE_ADD_GROUP],
      ])
      ],
      [
        'label' => 'admin.menu.researcher.label',
        'roles' => [Roles::ROLE_ADD_USER, Roles::ROLE_ADD_GROUP],
        'iconClass' => 'ti ti-user'
      ]);
    
    //Add the copy module menu items
    $this->elements[] = new MenuActionElement([], [
      'label' => 'admin.menu.copy.label',
      'route' => 'admin_copy',
      'iconClass' => 'fa fa-language fa-3x',
      'roles' => [ Roles::ROLE_COPY_MANAGER ],
    ]);
}
```

The `BaseAdminMenuExtension` also provides a function `addV1MenuItems` which allows you to add menu items defined in yaml to the menu service. Note that you may need to do some filtering on the items to prevent duplicates from being added.
