<?php

namespace Tests\Bazookas\APIFrameworkBundle\Services\Validation;

use Bazookas\APIFrameworkBundle\Exception\InvalidArgumentException;
use Bazookas\APIFrameworkBundle\Services\Validation\ValidationService;
use Exception;
use Symfony\Bundle\FrameworkBundle\Test\KernelTestCase;
use Symfony\Component\HttpFoundation\ParameterBag;
use Symfony\Component\Validator\Constraints\DateTime;
use Symfony\Component\Validator\Constraints\Email;
use Symfony\Component\Validator\Constraints\GreaterThan;
use Symfony\Component\Validator\Constraints\NotBlank;
use Symfony\Component\Validator\Constraints\Regex;
use Symfony\Component\Validator\Constraints\Type;

class ValidationServiceTest extends KernelTestCase
{
  private $container;

  public function setUp()
  {
    self::bootKernel();
    $this->container = self::$kernel->getContainer();
  }

  /**
   * Test the use of a single constraint
   */
  public function testConstraint() {
    $validator = new ValidationService($this->container->get('validator'));

    //Assert that a single constraint is validated correctly
    $requirement = [
      'constraint' => new Regex('/^\d+$/'),
      'required' => true
    ];

    $result = $validator->validate('test', 10, $requirement);
    $this->assertEquals($result, 10);
  }

  public function testSingularConstraintException() {
    $validator = new ValidationService($this->container->get('validator'));

    //Assert that an exception is thrown for an incorrect constraint
    $requirement = [
      'constraint' => new Regex('/^\d+$/'),
      'required' => true
    ];

    $asserted = false;
    try {
      $validator->validate('test', 'a', $requirement);
    } catch(Exception $e) {
      $this->assertTrue($e instanceof InvalidArgumentException);
      $asserted = true;
    }

    //should throw an error if exception is not thrown
    if (!$asserted) {
      $this->assertTrue(false, 'Expected InvalidArgumentException');
    }
  }

  /**
   * Test the use of multiple constraints
   */
  public function testMultipleConstraints() {
    $validator = new ValidationService($this->container->get('validator'));

    //assert that the validator works with two constraints
    $requirement = [
      'constraint' => [new Regex('/^\d+$/'), new GreaterThan(9)],
      'required' => true
    ];

    $result = $validator->validate('test', 10, $requirement);
    $this->assertEquals($result, 10);
  }

  public function testMultipleConstraintsException() {
    $validator = new ValidationService($this->container->get('validator'));

    //assert that the validators throws an exception with two constraints
    $requirement = [
      'constraint' => [new Regex('/^\d+$/'), new GreaterThan(9)],
      'required' => true
    ];

    $asserted = false;
    try {
      $validator->validate('test', 9, $requirement);
    } catch(Exception $e) {
      $asserted = true;
      $this->assertTrue($e instanceof InvalidArgumentException);
    }

    //should throw an error if exception is not thrown
    if (!$asserted) {
      $this->assertTrue(false, 'Expected InvalidArgumentException');
    }
  }

  /**
   * Test the required
   */
  public function testRequired() {
    $validator = new ValidationService($this->container->get('validator'));

    //Assert that a single constraint is validated correctly
    $requirement = [
      'constraint' => new Regex('/^\d+$/'),
      'required' => false
    ];

    $values = [null, 0, '', 3, ''];
    foreach($values as $value) {
      $result = $validator->validate('test', $value, $requirement);
      $this->assertEquals($result, $value);
    }
  }

  /**
   * Test the data conversion feature with a single constraint
   */
  public function testDataConversionSingleConstraint() {
    $validator = new ValidationService($this->container->get('validator'));

    $requirement = [
      'constraint' => new DateTime(),
      'required' => false,
      'convertToType' => ValidationService::TYPE_DATETIME,
      'conversionFormat' => 'U'
    ];

    $time = new \DateTime();
    $result = $validator->validate('test', $time->getTimestamp(), $requirement);

    $this->assertTrue($result instanceof \DateTime);
    $this->assertEquals($result->getTimestamp(), $time->getTimestamp());
  }

  /**
   * Test the data conversion feature with a multiple constraints
   */
  public function testDataConversionMultipleConstraints() {
    $validator = new ValidationService($this->container->get('validator'));

    $past = new \DateTime();
    $past->modify('-10 seconds');

    $requirement = [
      'constraint' => [new DateTime(), new GreaterThan($past)],
      'required' => false,
      'convertToType' => ValidationService::TYPE_DATETIME,
      'conversionFormat' => 'U'
    ];

    $time = new \DateTime();
    $result = $validator->validate('test', $time->getTimestamp(), $requirement, true);

    $this->assertTrue($result instanceof \DateTime);
    $this->assertEquals($result->getTimestamp(), $time->getTimestamp());
  }

  private function performObjectTestSetup() {
    $validator = new ValidationService($this->container->get('validator'));

    $requirements = [
      'number' => [
        'constraint' => new Regex('/^\d+$/'),
        'required' => false,
        'dependencies' => ['number2', 'string2']
      ],
      'number2' => [
        'constraint' => new Regex('/^\d+$/'),
        'required' => true
      ],
      'string' => [
        'constraint' => new NotBlank(),
        'required' => false
      ],
      'string2' => [
        'constraint' => new NotBlank(),
        'required' => false
      ],
      'email' => [
        'constraint' => [new Type('string'), new Email()],
        'required' => false,
        'dependencies' => 'email2'
      ],
      'email2' => [
        'constraint' => [new NotBlank(), new Email()],
        'required' => true
      ],
      //add a test for an empty constraint
      'empty' => []
    ];

    $data = [
      'number' => 0,
      'number2' => 12,
      'string' => null,
      'string2' => 'a test',
      'email' => '',
      'email2' => 'simon@bazookas.be',
      'empty' => 3
    ];

    return [$validator, $requirements, $data];
  }

  /**
   * Tests the validate data function (which can validate entire objects)
   */
  public function testValidateDataParams() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    $result = $validator->validateData($requirements, $data);
    $this->assertTrue($result instanceof ParameterBag);

    foreach($data as $key => $item) {
      $this->assertEquals($result->get($key), $item);
    }
  }

  public function testValidateDataDependencies() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    $result = $validator->validateData($requirements, $data);
    $this->assertTrue($result instanceof ParameterBag);

    foreach($data as $key => $item) {
      $this->assertEquals($result->get($key), $item);
    }
  }

  public function testValidateDataDependenciesException() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    $requirements['number'] = [
      'constraint' => new Regex('/^\d+$/'),
      'required' => false,
      'dependencies' => ['number2', 'string']
    ];

    unset($data['email2']);

    $asserted = false;
    try {
      $validator->validateData($requirements, $data);
    } catch(Exception $e) {
      $asserted = true;
      $this->assertTrue($e instanceof InvalidArgumentException);
    }

    //should throw an error if exception is not thrown
    if (!$asserted) {
      $this->assertTrue(false, 'Expected InvalidArgumentException');
    }
  }

  public function testValidateOptionalDependencies() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    unset($data['number']);
    unset($data['string2']);

    $result = $validator->validateData($requirements, $data);
    $this->assertTrue($result instanceof ParameterBag);

    foreach($data as $key => $item) {
      $this->assertEquals($result->get($key), $item);
    }
  }

  public function testValidateOptionalDependenciesException() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    unset($data['string2']);

    $asserted = false;
    try {
      $validator->validateData($requirements, $data);
    } catch(Exception $e) {
      $asserted = true;
      $this->assertTrue($e instanceof InvalidArgumentException);
    }

    //should throw an error if exception is not thrown
    if (!$asserted) {
      $this->assertTrue(false, 'Expected InvalidArgumentException');
    }
  }

  public function testValidateDataDependenciesException2() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    unset($data['email2']);

    $asserted = false;
    try {
      $validator->validateData($requirements, $data);
    } catch(Exception $e) {
      $asserted = true;
      $this->assertTrue($e instanceof InvalidArgumentException);
    }

    //should throw an error if exception is not thrown
    if (!$asserted) {
      $this->assertTrue(false, 'Expected InvalidArgumentException');
    }
  }

  public function testValidateDataExclusions() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    $requirements['email'] = [
      'constraint' => [new Type('string'), new Email()],
      'required' => false,
      'excludes' => 'email3',
      'dependencies' => 'email2'
    ];

    $result = $validator->validateData($requirements, $data);
    $this->assertTrue($result instanceof ParameterBag);

    foreach($data as $key => $item) {
      $this->assertEquals($result->get($key), $item);
    }
  }

  public function testValidateDataExclusionsException() {
    list($validator, $requirements, $data) = $this->performObjectTestSetup();

    $requirements['email'] = [
      'constraint' => [new Type('string'), new Email()],
      'required' => false,
      'excludes' => 'email2',
      'dependencies' => 'email2'
    ];

    $asserted = false;
    try {
      $validator->validateData($requirements, $data);
    } catch(Exception $e) {
      $asserted = true;
      $this->assertTrue($e instanceof InvalidArgumentException);
    }

    //should throw an error if exception is not thrown
    if (!$asserted) {
      $this->assertTrue(false, 'Expected InvalidArgumentException');
    }
  }
}
