<?php

namespace Bazookas\CommonBundle\Entity\Base;

use Bazookas\CommonBundle\Entity\Interfaces\EntityInterface;
use Bazookas\CommonBundle\Exception\NotImplementedException;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\Mapping as ORM;

/**
 * @ORM\MappedSuperclass()
 */
abstract class BaseNonBinaryTreeNode extends BaseEntity {

  const DEPTH_FIRST_TRAVERSAL = 'DEPTH_FIRST_TRAVERSAL';
  const BREADTH_FIRST_TRAVERSAL = 'BREADTH_FIRST_TRAVERSAL';

  /**
   * @var BaseNonBinaryTreeNode
   */
  protected $parent;

  /**
   * @var BaseNonBinaryTreeNode[]
   */
  protected $children;

  /**
   * @var mixed
   */
  protected $value;

  public function __construct() {
    parent::__construct();

    $this->children = new ArrayCollection();
  }

  /**
   * @return BaseNonBinaryTreeNode
   */
  public function getParent()
  {
    return $this->parent;
  }

  /**
   * @param BaseNonBinaryTreeNode $parent
   * @return $this
   */
  public function setParent(BaseNonBinaryTreeNode $parent)
  {
    $this->parent = $parent;

    return $this;
  }

  /**
   * @return BaseNonBinaryTreeNode[]
   */
  public function getChildren()
  {
    return $this->children;
  }

  /**
   * @param BaseNonBinaryTreeNode $newNode
   * @return $this
   */
  public function addChild(BaseNonBinaryTreeNode $newNode) {
    if (!$this->children->contains($newNode)) {
      $newNode->setParent($this);
      $this->children->add($newNode);
    }

    return $this;
  }

  /**
   * @param BaseNonBinaryTreeNode $newNode
   * @return $this
   */
  public function removeChild(BaseNonBinaryTreeNode $newNode) {
    if ($this->children->contains($newNode)) {
      $this->children->removeElement($newNode);
    }

    return $this;
  }

  /**
   * @param BaseNonBinaryTreeNode[] $children
   * @return $this
   */
  public function setChildren(array $children)
  {
    $this->children = $children;

    return $this;
  }

  /**
   * @return mixed
   */
  public function getValue()
  {
    return $this->value;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function setValue($value)
  {
    $this->value = $value;

    return $this;
  }

  /**
   * @param BaseNonBinaryTreeNode $newNode
   * @param $parentValue
   * @param string $algorithm
   * @return $this
   * @throws \Exception
   */
  public function insert(BaseNonBinaryTreeNode &$newNode, $parentValue, $algorithm = self::DEPTH_FIRST_TRAVERSAL ) {
    //Does the new node already exist?
    $node = $this->search($newNode->getValue(), $algorithm);
    if ($node !== false) {
      $parent = $node->getParent();

      //We only accept that the new node already exists if its value is already in the tree and its parent has the same value as the value given.
      if ($parent->getValue() === $parentValue) {
        return $this;
      }
    }

    $parent = $this->search($parentValue, $algorithm);

    if ($parent !== false) {
      $parent->addChild($newNode);
    } else {
      throw new \Exception('Unknown parent with value');
    }

    return $this;
  }

  /**
   * @param mixed $value
   * @param string $algorithm
   * @return BaseNonBinaryTreeNode|false
   * @throws \Exception
   */
  public function search($value, $algorithm = self::DEPTH_FIRST_TRAVERSAL ) {
    switch($algorithm) {
      case self::BREADTH_FIRST_TRAVERSAL:
        return $this->breadthFirstTraversal([$this, 'searchForValue'], [$value]);
      case self::DEPTH_FIRST_TRAVERSAL:
        return $this->depthFirstTraversal([$this, 'searchForValue'], [$value]);
      default:
        throw new \Exception('Unknown traversal algorithm specified.');
    }
  }

  /**
   * Traverse the tree in breadth-first fashion and output the value. Use this function for debugging purposes
   * @param string $algorithm
   * @throws \Exception
   */
  public function printTree($algorithm = self::DEPTH_FIRST_TRAVERSAL) {
    switch($algorithm) {
      case self::DEPTH_FIRST_TRAVERSAL:
        $this->depthFirstTraversal([$this, 'printNode']);
        break;
      case self::BREADTH_FIRST_TRAVERSAL:
        $this->breadthFirstTraversal([$this, 'printNode']);
        break;
      default:
        throw new \Exception('Unknown traversal algorithm specified.');
    }
  }

  /* ------------------------------------------------------------------------
   * Traversal helper functions
   * ------------------------------------------------------------------------
   */

  /**
   * @see https://en.wikipedia.org/wiki/Breadth-first_search
   * @param callable $callable the function that should be executed on each traversed node
   * @param array $params the parameters to be passed to the callable, the first element will always be the current node
   * @return BaseNonBinaryTreeNode|false
   */
  public function breadthFirstTraversal(callable $callable, array $params = []) {
    $queue = new \SplQueue();
    $discovered = [];

    $queue->enqueue($this);
    $discovered[spl_object_hash($this)] = true;

    while (!$queue->isEmpty()) {
      $currentNode = $queue->dequeue();

      //Do something with the value
      array_unshift($params, $currentNode);
      $returnValue = call_user_func_array($callable, $params);
      if ($returnValue !== false) {
        return $returnValue;
      }

      $children = $currentNode->getChildren();
      foreach($children as $childNode) {
        $objectHash = spl_object_hash($childNode);
        if (!isset($discovered[$objectHash])) {
          $discovered[$objectHash] = true;
          $queue->enqueue($childNode);
        }
      }
    }

    return false;
  }

  /**
   * @see https://en.wikipedia.org/wiki/Depth-first_search
   * @param callable $callable the function that should be executed on each traversed node
   * @param array $params the parameters to be passed to the callable, the first element will always be the current node
   * @return BaseNonBinaryTreeNode|false
   */
  public function depthFirstTraversal(callable $callable, array $params = []) {
    $stack = new \SplStack();
    $discovered = [];

    $stack->push($this);
    $discovered[spl_object_hash($this)] = true;

    while (!$stack->isEmpty()) {
      $currentNode = $stack->pop();

      //Do something with the value
      array_unshift($params, $currentNode);
      $returnValue = call_user_func_array($callable, $params);
      if ($returnValue !== false) {
        return $returnValue;
      }

      $children = $currentNode->getChildren();
      foreach($children as $childNode) {
        $objectHash = spl_object_hash($childNode);
        if (!isset($discovered[$objectHash])) {
          $discovered[$objectHash] = true;
          $stack->push($childNode);
        }
      }
    }

    return false;
  }

  /**
   * Helper function to search for a particular value and return from the traversal once that value is found.
   * @param BaseNonBinaryTreeNode $currentNode
   * @param $value
   * @return BaseNonBinaryTreeNode|bool
   */
  protected function searchForValue(BaseNonBinaryTreeNode $currentNode, $value) {
    if ($currentNode->getValue() === $value) {
      return $currentNode;
    }

    return false;
  }

  /**
   * Helper function to print an entire tree.
   * @param BaseNonBinaryTreeNode $currentNode
   */
  protected function printNode(BaseNonBinaryTreeNode $currentNode) {
    $currentValue = $currentNode->getValue();
    if ($currentValue instanceof EntityInterface) {
      //Use echo as dump risks us running into the recursive object limit of php
      echo $currentValue->getId() . ' ';
    } else {
      //Use echo as dump risks us running into the recursive object limit of php
      echo $currentValue . ' ';
    }
  }

  /**
   * How do we determine if two items are equal?
   * @param BaseNonBinaryTreeNode $a
   * @param BaseNonBinaryTreeNode $b
   * @return bool
   */
  protected function compare(BaseNonBinaryTreeNode $a, BaseNonBinaryTreeNode $b) {
    return $a->getValue() === $b->getValue();
  }
}
