<?php

namespace Bazookas\CommonBundle\Repository\Base;

use Bazookas\CommonBundle\Entity\Interfaces\EntityInterface;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\EntityManagerInterface;
use Bazookas\CommonBundle\Repository\Interfaces\RepositoryInterface;
use Bazookas\CommonBundle\Entity\Interfaces\EntityDetailParentInterface;
use Bazookas\CommonBundle\Entity\Interfaces\LocalisedEntityInterface;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Query;
use Doctrine\ORM\QueryBuilder;
use Doctrine\ORM\Tools\Pagination\Paginator;

abstract class BaseRepository extends EntityRepository implements RepositoryInterface {

  /**
   * @deprecated to be removed in v2, use $this->_em instead
   * @var EntityManagerInterface
   */
  protected $entityManager;

  public function __construct(EntityManagerInterface $entityManager, ClassMetadata $classMetadata = null) {
    if (empty($classMetadata)) {
      $classMetadata = $entityManager->getClassMetadata($this->getEntityClass());
    }

    parent::__construct($entityManager, $classMetadata);
    $this->entityManager = $this->_em;
  }

  /**
   * @return string
   */
  public function getEntityClass() {
    return $this->_entityName;
  }

  /**
   * Get all of the results for which the modified date time is after $lastModified
   * @param \DateTime|null $lastModified
   * @param string $entityClass
   * @param string $resultFormat
   * @return array
   */
  public function getResultsByLastModified(\DateTime $lastModified = null, string $entityClass = '' , string $resultFormat = 'array') {
    if (empty($entityClass)) {
      $entityClass = $this->_entityName;
    }

    $qb = $this->_em->createQueryBuilder();
    $qb
      ->select('e')
      ->from($entityClass, 'e')
    ;

    if (!empty($lastModified)) {
      $qb
        ->andWhere($qb->expr()->gt('e.modified', ':modified'))
        ->setParameter('modified', $lastModified)
      ;
    }

    if ($resultFormat === 'array') {
      return $qb->getQuery()->getArrayResult();
    } else {
      return $qb->getQuery()->getResult();
    }
  }

  /**
   * Get all of the results for which the modified date time is after $lastModified. This query also filters results based on the passed $locale
   * @param $locale
   * @param \DateTime|null $lastModified
   * @param string $entityClass
   * @param string $resultFormat
   * @return array
   */
  public function getLocalisedResultByLastModified(string $locale, \DateTime $lastModified = null, string $entityClass = '' , string $resultFormat = 'array') {
    if (empty($entityClass)) {
      $entityClass = $this->_entityName;
    }

    $qb = $this->_em->createQueryBuilder();
    $qb
      ->select('e')
      ->from($entityClass, 'e')
    ;

    $classImplements = class_implements($entityClass);
    $classImplements = is_array($classImplements) ? $classImplements : [$classImplements];

    if (in_array(EntityDetailParentInterface::class, $classImplements)) {
      //the class has a localised details item
      $qb
        ->addSelect('ed')
        ->join('e.details', 'ed', 'WITH', $qb->expr()->eq('ed.locale', ':locale'))
      ;

      if (!empty($lastModified)) {
        $qb
          ->andWhere(
            $qb->expr()->orX(
              $qb->expr()->gt('e.modified', ':modified'),
              $qb->expr()->gt('ed.modified', ':modified')
            )
            )
          ->setParameter('modified', $lastModified)
        ;
      }
    } else if (in_array(LocalisedEntityInterface::class, $classImplements)) {
      //the class itself is a localised item
      $qb
        ->andWhere($qb->expr()->eq('e.locale', ':locale'))
      ;

      if (!empty($lastModified)) {
        $qb
          ->andWhere($qb->expr()->gt('e.modified', ':modified'))
          ->setParameter('modified', $lastModified)
        ;
      }

    } else {
      throw new \InvalidArgumentException('The given entity does not implement: '.EntityDetailParentInterface::class.' or '.LocalisedEntityInterface::class);
    }

    $qb->setParameter('locale', $locale);

    if ($resultFormat === 'array') {
      return $qb->getQuery()->getArrayResult();
    } else {
      return $qb->getQuery()->getResult();
    }
  }

  /**
   * @param array $ids
   * @param int $hydrationMode
   * @return array
   */
  public function getEntitiesById(array $ids, $hydrationMode = Query::HYDRATE_OBJECT) {
    $qb = $this->_em->createQueryBuilder();
    $qb
      ->select('e')
      ->from($this->_entityName, 'e')
      ->andWhere($qb->expr()->in('e.id', ':ids'))
      ->setParameter('ids', $ids)
    ;

    return $qb->getQuery()->getResult($hydrationMode);
  }

  /**
   * @deprecated due to typo, use updateEntities instead, to be removed in v2
   * @param array $entities
   */
  public function updateEntites(array $entities) {
    $this->updateEntities($entities);
  }

  /**
   * @param array $entities
   */
  public function updateEntities(array $entities) {
    foreach($entities as $entity) {
      $this->_em->persist($entity);
    }

    $this->_em->flush();
  }

  /**
   * @param array $entities
   */
  public function refreshEntities(array $entities) {
    foreach($entities as $entity) {
      $this->_em->refresh($entity);
    }
  }

  /**
   * @param array $entities
   */
  public function removeEntities(array $entities){
    foreach($entities as $entity) {
      $this->_em->remove($entity);
    }

    $this->_em->flush();
  }

  /**
   * @param QueryBuilder $qb
   * @param int $page
   * @param int $limit
   * @param int $hydrationMode
   * @return array|EntityInterface[]
   */
  public function getPageableResult(QueryBuilder $qb, int $page, int $limit = 50, $hydrationMode = Query::HYDRATE_ARRAY) {
    $qb
      ->setFirstResult($page * $limit)
      ->setMaxResults($limit)
    ;

    $query = $qb->getQuery();
    $query->setHydrationMode($hydrationMode);

    $paginator = new Paginator($query, true);
    $pageCount = intval(ceil($paginator->count() / $limit));

    $results = [];
    foreach($paginator as $page) {
      $results[] = $page;
    }

    return ['results' => $results, 'pageCount' => $pageCount];
  }

  /**
   * @param string $entityClass the class of the entity to be queried
   * @param string $groupIdentifierField the field by which the entity should be grouped
   * @param string $maxField the field which should be maximised
   * @return \Doctrine\ORM\QueryBuilder qb
   */
  public function createGreatestRowPerGroupQueryBuilder(string $entityClass, string $groupIdentifierField, string $maxField) {
    //witchcraft explained: http://stackoverflow.com/questions/7745609/sql-select-only-rows-with-max-value-on-a-column
    $qb = $this->_em->createQueryBuilder();
    $qb
      ->select('e1')
      ->from($entityClass, 'e1')

      ->leftJoin($entityClass, 'e2' ,'WITH', $qb->expr()->andX(
        $qb->expr()->eq('e1.'. $groupIdentifierField, 'e2.'.$groupIdentifierField),
        $qb->expr()->lt('e1.'.$maxField, 'e2.'.$maxField)
      ))
      ->andWhere($qb->expr()->isNull('e2.id'))
    ;

    return $qb;
  }
}
